#include <Application.h>
#include <Bitmap.h>
#include <Alert.h>
#include <Box.h>
#include <Message.h>
#include <MenuItem.h>
#include <stdlib.h>
#include "SupportKit.h"
#include "LoginBox.h"
#include "MiscStuff.h"
#include "constants.h"
#include "Globals.h"
#include "Say.h"

//=====================================================================

LoginWindow::LoginWindow(BRect frame)
 				: BWindow(frame, "Login", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE )
{
	BBitmap* mainlogo = NULL;
	
	// Set up the views
	BRect aRect( Bounds() );
	logView = new LoginView(aRect);
	aRect.bottom = 150;
	progressView = new LoginProgressView(aRect);
	progressView->Hide();
	countView = new LoginCountdownView( BRect( 5,206,196,237) );
	
	GetBitmapFromResources( mainlogo, 42 );
	BRect bframe = Bounds();
	bframe.bottom = 77;
	bview = new BitmapView( bframe, mainlogo );

	// Now add them to the window
	AddChild( bview );
	AddChild( logView );
	AddChild( progressView );
	logView->AddChild( countView );
	countView->Hide();
	logView->ScreenName->MakeFocus(true);
	
	// load the login settings
	LoadLoginSettings();
}

//---------------------------------------------------------------------

LoginWindow::~LoginWindow() {
}

//---------------------------------------------------------------------

bool LoginWindow::QuitRequested()
{
	SaveLoginSettings( false );
	if( !user->LoggedIn() )
		be_app->PostMessage(B_QUIT_REQUESTED);
	return(true);
}

//---------------------------------------------------------------------

void LoginWindow::MessageReceived(BMessage* message)
{
	switch(message->what)
	{
		case BEAIM_SIGN_ON:
			// added to not try to login while it's already happening
			countView->StopCountdown();
			DoLogin();
			break;
			
		case BEAIM_CANCEL_SIGN_ON:
			PostAppMessage( new BMessage(BEAIM_CANCEL_SIGN_ON) );
			SetProgressMode( false );
			break;
			
		case BEAIM_SET_LOGIN_STEP_COUNT:	// set the number of steps in the progress bar
			progressView->Progress->SetMaxValue( message->FindFloat("count") );
			break;
			
		case BEAIM_LOGIN_STEP:
			progressView->Progress->Update( 1 );	// move the indicator on the progress bar
			if( message->FindBool("update_text") )
				progressView->Caption->SetText( message->FindString("status") );
			break;

		case BEAIM_LOGIN_FAILURE: {
			// turn off the deskbar icon
			BMessage* noDeskbarIcon = new BMessage(BEAIM_TOGGLE_DESKBAR_ICON);
			noDeskbarIcon->AddBool( "enabled", false );
			PostAppMessage(noDeskbarIcon);
			
			// fix the progress bar, and display the reason
			SetProgressMode( false );
			Say( (char*)message->FindString("reason") );
			break;
		}
			
		case BEAIM_OPEN_PREFS: {
			BMessage* msg = new BMessage(BEAIM_OPEN_SINGLE_WINDOW);
			msg->AddInt32( "wtype", SW_PREFERENCES );
			msg->AddBool( "globalonly", true );
			msg->AddInt32( "posnum", 1 );
			PostAppMessage( msg );
			break;
		}
			
		case PEOPLE_SELECTOR_MSG:
			logView->LoadUserSettings( message->FindInt32("index") );
			break;
			
		case LOGIN_COUNTDOWN_MSG:
			countView->Tick();
			break;
		
		case LOGIN_STOP_COUNTDOWN:
			countView->StopCountdown();
			break;
			
		case LOGIN_EDIT_PW:		
		case LOGIN_SAVE_PW_CHECKBOX:
			logView->AdjustPreLoginSettings();
			break;

		case B_CANCEL:
			QuitRequested();
			Close();
			break;
						
		default:
			BWindow::MessageReceived(message);
	}
}

//---------------------------------------------------------------------

// the cancel function
void LoginWindow::DispatchMessage( BMessage* msg, BHandler* handler ) {

	// if it's a cancel key, post a B_CANCEL message
	if( msg->what == B_KEY_DOWN )
		if( msg->FindString("bytes")[0] == B_ESCAPE ) {
			PostMessage( new BMessage(B_CANCEL) );
			return;
		}
	
	// our work here is done... dispatch normally
	BWindow::DispatchMessage( msg, handler );
}

//---------------------------------------------------------------------

void LoginWindow::SetCountdownMode( bool mode ) {

	// remake the window to show the countdown display
	if( mode ) {
		logView->ResizeTo( 200, 240 );
		ResizeTo( 200, 240 );
		countView->Show();
		logView->EnableControls(false);
		countView->cancelButton->MakeFocus(true);
		countView->DoCountdown();
	}
	
	// turn off the countdown display and stop the countdown
	else { 
		logView->ResizeTo( 200, 205 );
		ResizeTo( 200, 205 );
		countView->Hide();
		logView->EnableControls(true);
		logView->ScreenName->MakeFocus(true);
	}
}

//---------------------------------------------------------------------

void LoginWindow::SetProgressMode( bool mode ) {

	// remake the window to show the countdown display
	if( mode ) {
		ResizeTo( 200, 150 );
		while( !logView->IsHidden() )
			logView->Hide();
		while( progressView->IsHidden() )
			progressView->Show();
		progressView->Progress->Update( -(progressView->Progress->CurrentValue()) );
		progressView->CancelButton->MakeFocus(true);
		progressView->CancelButton->MakeDefault(true);
	}

	// turn off the countdown display and stop the countdown
	else { 
		logView->EnableControls(true);
		while( logView->IsHidden() )
			logView->Show();
		logView->ResizeTo( 200, 205 );
		ResizeTo( 200, 205 );
		while( !progressView->IsHidden() )
			progressView->Hide();
		logView->ScreenName->MakeFocus(true);
		logView->SignOnButton->MakeDefault(true);
		user->SetLoggedIn( false );
	}
}

//---------------------------------------------------------------------

void LoginWindow::LoadLoginSettings() {

	// grab the name of the person who last logged in...
	// and if there was one, load their settings
	if( user->LastLogin().Length() )
		logView->LoadUserSettings( const_cast<char*>(user->LastLogin().String()) );
		
	// Now... if the auto-login check box is checked, then do the autologin
	if( logView->AutoLogin->Value() ) {
		if( logView->Password->actualText() != "" )
			SetCountdownMode( true );
		else
			logView->Password->MakeFocus(true);
	}
}

//---------------------------------------------------------------------

void LoginWindow::SaveLoginSettings( bool saveNewUser ) {

	BString userName;
	
	// save the window position
	BRect winPos = Frame();
	user->WriteInt32( "login.x", int32(winPos.LeftTop().x), true );
	user->WriteInt32( "login.y", int32(winPos.LeftTop().y), true );

	// Get and save the username, and let the user view save its own stuff
	userName = logView->ScreenName->Text();
	user->SetLastLogin( userName );
	logView->SaveUserSettings( const_cast<char*>(userName.String()), saveNewUser );
	user->WriteGlobalFile();
}

//---------------------------------------------------------------------

void LoginWindow::DoLogin() {

	BMessage* sendMessage = NULL;

	// setup the window for the login
	SetProgressMode( true );
	
	// set the user and save the login settings
	user->SetUser( logView->ScreenName->Text() );
	SaveLoginSettings( true );

	// send the login message
	sendMessage = new BMessage(BEAIM_SIGN_ON);
	sendMessage->AddString( "userid", logView->ScreenName->Text() );
	sendMessage->AddString( "password", (char*)logView->Password->actualText().String() );
	PostAppMessage( sendMessage );
}

//=====================================================================

LoginView::LoginView(BRect rect)
	   	   : BView(rect, "login_view", B_FOLLOW_LEFT | B_FOLLOW_TOP, B_WILL_DRAW)
{
	SetViewColor( 216, 216, 216 );

	// Construct the views
	BStringView* tempSV = new BStringView( BRect(0,70,198,85), "", versionString.String() );
	tempSV->SetFont(be_plain_font);
	tempSV->SetFontSize(9);
	tempSV->SetAlignment(B_ALIGN_RIGHT);
	ScreenName = new BTextControl( BRect(8,90,175,110), "screenname", "Screen Name:", "", NULL );
	ScreenName->SetViewColor( 216, 216, 216 );
	ScreenName->SetDivider( 75 );
	Password = new PassControl( BRect(8,115,192,135), "password", "Password:", "", new BMessage(31337) );
	Password->SetViewColor( 216, 216, 216 );
	Password->SetDivider( 75 );
	Password->SetModificationMessage( new BMessage(LOGIN_EDIT_PW) );
	SavePassword = new BCheckBox( BRect(8,143,103,163), "Save Password", "Save Password", new BMessage(LOGIN_SAVE_PW_CHECKBOX) );
	SavePassword->SetViewColor( 216, 216, 216 );
	AutoLogin = new BCheckBox( BRect(126,143,219,163), "autologin", "Auto-login", new BMessage(LOGIN_AUTOLOGIN_CHECKBOX) );
	AutoLogin->SetViewColor( 216, 216, 216 );
	SignOnButton = new BButton(BRect(132,172,192,192), "loginbutton", "Sign On", new BMessage(BEAIM_SIGN_ON)); 
	SignOnButton->MakeDefault( true );
	SetupButton = new BButton(BRect(8,172,55,192), "loginbutton", "Setup", new BMessage(BEAIM_OPEN_PREFS));
	MakeUserList();	
	userList = new BMenuField( BRect(178,88,193,107), NULL, NULL, userMenu, true );

	// Now add them to the window
	AddChild( ScreenName );
	AddChild( userList );
	AddChild( Password );
	AddChild( SavePassword );
	AddChild( AutoLogin );
	AddChild( SetupButton );
	AddChild( SignOnButton );
	AddChild( tempSV );
	
	// disable some stuff
	AdjustPreLoginSettings();
}

//---------------------------------------------------------------------

void LoginView::LoadUserSettings( int32 index ) {

	char userName[DISPLAY_NAME_MAX];

	// Grab the user's name and call the other settings function
	strcpy( userName, (char*)userList->Menu()->ItemAt(index)->Label() );
	LoadUserSettings( userName );
}

//---------------------------------------------------------------------

void LoginView::LoadUserSettings( char* userName ) {

	BString password;
	bool autoLogin = false, savePassword = false;

	// save the old stuff and grab the user's info
	SaveUserSettings( const_cast<char*>(ScreenName->Text()) );
	user->GetUserLoginParameters( userName, password, autoLogin, savePassword );

	// set stuff up based on the info in the user file
	SavePassword->SetValue( savePassword );
	AutoLogin->SetValue( autoLogin );
	ScreenName->SetText( userName );
	if( savePassword && password.Length() ) {
		Password->SetText( password.String() );
	} else {
		Password->SetText("");
	}
	
	// enable/disable stuff
	AdjustPreLoginSettings();
}

//---------------------------------------------------------------------

void LoginView::SaveUserSettings( char* userName, bool saveNewUser ) {

	// sanity check
	if( !strlen(userName) )
		return;
		
	// save 'em
	user->SaveLoginParameters( BString(userName),
							   Password->actualText(),
							   AutoLogin->Value(),
							   SavePassword->Value(),
							   saveNewUser );
}

//---------------------------------------------------------------------

void LoginView::EnableControls( bool enabled ) {

	ScreenName->SetEnabled( enabled );
	Password->SetEnabled( enabled );
	SavePassword->SetEnabled( enabled );
	AutoLogin->SetEnabled( enabled );
	userList->SetEnabled( enabled );
	SetupButton->SetEnabled( enabled );
	AdjustPreLoginSettings();
}

//---------------------------------------------------------------------

void LoginView::MakeUserList() {

	BString name;
	bool ret = true;
	bool start = true;
	int32 msgID = 0;
	BMessage* selectMessage;
	userMenu = new BMenu("");

	// add the users
	while( ret ) {
		ret = user->GetNextUser( name, start );
		if( !ret )
			break;		
		if( start )
			start = false;
		if( !name.Length() )
			continue;
		selectMessage = new BMessage(PEOPLE_SELECTOR_MSG);
		selectMessage->AddInt32( "index", msgID++ );
		userMenu->AddItem( new BMenuItem(name.LockBuffer(0), selectMessage) );
		name.UnlockBuffer();
	}
	
	// no people? Add a "(none)" item
	if( msgID == 0 ) {
		BMenuItem* disabledItem = new BMenuItem("(none)", NULL);
		disabledItem->SetEnabled( false );
		userMenu->AddItem( disabledItem );
	}
}

//---------------------------------------------------------------------

// handle all the enable/disable silliness for the password, save password, and autologin controls
void LoginView::AdjustPreLoginSettings() {

	//char duhski[100];
	//sprintf( duhski, "%ld", Password->actualText().Length() );
	//Window()->SetTitle( duhski );

	// can't do much in terms of autologging or whatever without a password
	if( Password->actualText().Length() == 0 ) {
		if( SavePassword->IsEnabled() ) {
			SavePassword->SetEnabled(false);
			SignOnButton->SetEnabled(false);
			AutoLogin->SetEnabled(false);
		}
		return;
	}
	
	// quick hack...
	if( !Password->IsEnabled() )
		return;
	
	// otherwise, make sure that at least Save Password is enabled
	if( !SavePassword->IsEnabled() )
		SavePassword->SetEnabled(true);
	if( !SignOnButton->IsEnabled() )
		SignOnButton->SetEnabled(true);
			
	// and only allow AutoLogin to be enabled if SavePassword is
	if( (bool)SavePassword->Value() != AutoLogin->IsEnabled() )
		AutoLogin->SetEnabled( (bool)SavePassword->Value() );
}

//=====================================================================

LoginProgressView::LoginProgressView(BRect rect)
	   	   : BView(rect, "progress_view", B_FOLLOW_LEFT | B_FOLLOW_TOP, B_WILL_DRAW)
{
	SetViewColor( 216, 216, 216 );
	Progress = new BStatusBar( BRect(5,67,195,87), "progress" );
	Progress->SetMaxValue(11);
	Caption = new BStringView( BRect(5,92,195,112), "caption", "" );	
	CancelButton = new BButton( BRect(75,120,125,140), "cancel", "Cancel", new BMessage(BEAIM_CANCEL_SIGN_ON) );
	
	AddChild( Caption );
	AddChild( Progress );
	AddChild( CancelButton );
}

//=====================================================================

LoginCountdownView::LoginCountdownView(BRect rect)
	   	   : BView(rect, "count_view", B_FOLLOW_LEFT | B_FOLLOW_TOP, B_WILL_DRAW)
{
	SetViewColor( 216, 216, 216 );
	surroundBox = new BBox( Bounds() );
	cancelButton = new BButton( BRect(136,3,186,25), "cancel", "Cancel", new BMessage(LOGIN_STOP_COUNTDOWN) );
	Caption = new BStringView( BRect(7,4,195,24), "caption", "Auto-login in 5 seconds..." );
	
	AddChild( cancelButton );
	AddChild( surroundBox );
	AddChild( Caption );
}

//---------------------------------------------------------------------

void LoginCountdownView::DoCountdown() {

	// default is 3 seconds... should probably make an option for this
	counter = 3;

	// make the thread and do the first tick
	count_thread = spawn_thread( counter_thread_entry, "countdown_thread", B_LOW_PRIORITY, this );
	if( count_thread <= 0 )
		Say( "Error! Thread not spawned!" );
	else {
		resume_thread( count_thread );
		Tick();
	}
}

//---------------------------------------------------------------------

void LoginCountdownView::StopCountdown() {

	//beep();

	if( count_thread ) {
		kill_thread( count_thread );
		count_thread = 0;
	}
	
	// Remake the window to hide this view
	LoginWindow* window = dynamic_cast<LoginWindow*>( Window() );
	window->SetCountdownMode( false );
}

//---------------------------------------------------------------------

void LoginCountdownView::Tick() {

	char Label[50];
	LoginWindow* parent = dynamic_cast<LoginWindow*>(Window());
	
	if( counter > 0 ) {
		sprintf( Label, "Auto-login in %d second%s...", counter, counter==1?"":"s" );
		Caption->SetText( Label );
	} else {
	
		// Do the login
		kill_thread( count_thread );
		count_thread = 0;
		parent->DoLogin();
	}

	counter--;
}

//---------------------------------------------------------------------

int32 LoginCountdownView::counter_thread_entry( void *arg ) {
	LoginCountdownView* this_class = (LoginCountdownView*)arg;
	return this_class->do_count();
}

//---------------------------------------------------------------------

int32 LoginCountdownView::do_count() {

	// sleep for one second, then send a tick message
	while( true ) {
		snooze( 1000000 );
		Window()->PostMessage( new BMessage(LOGIN_COUNTDOWN_MSG) );
	}
}
		
//=====================================================================
