#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ListItem.h>
#include <SupportKit.h>
#include <Bitmap.h>
#include "PeopleData.h"
#include "BuddyList.h"
#include "MiscStuff.h"
#include "Globals.h"
#include "Say.h"
#include "constants.h"

//#define PLIST_DEBUG

BBitmap* PeopleItem::awayIcon = NULL;

//-------------------------------------------------------------------------

PeopleList::PeopleList() {
	display = NULL;
	setup = NULL;
	commitAdds = 0;
	
	// make the locker
	lock = new BLocker( "people_lock" );
}

//-------------------------------------------------------------------------

PeopleList::~PeopleList() {

	lock->Lock();

	// clear stuff out
	Clear();

	// delete the BLocker
	lock->Unlock();
	delete lock;
}

//-------------------------------------------------------------------------

void PeopleList::Clear() {

	lock->Lock();
	GroupInfo* group;
	
	display = setup = NULL;
	commitAdds = 0;

	// first the piddly stuff
	commitList.Clear();
	changeList.Clear();
	
	// Empty the main list
	while( main.Pop(group) )
		delete group;		
	
	lock->Unlock();
}

//-------------------------------------------------------------------------

void PeopleList::SetDisplayList( BOutlineListView* list ) {

	lock->Lock();
	display = list;
	GroupItem* groupItem;
	
	// Load all groups into the display
	for( unsigned int i = 0; i < main.Count(); ++i ) {
		groupItem = new GroupItem( main[i]->group, 0, main[i]->total );
		main[i]->item = groupItem;
		display->AddItem( groupItem );
	}		
	
	lock->Unlock();	
}

//-------------------------------------------------------------------------

void PeopleList::SetSetupList( BOutlineListView* list ) {

	lock->Lock();
	setup = list;
	PeopleSetupItem* groupItem;
	PeopleSetupItem* personItem;
	GenList<PersonInfo*>* plist;
	
	// Load all the groups into the list
	for( unsigned int i = 0; i < main.Count(); ++i ) {
	
		groupItem = new PeopleSetupItem( main[i]->group, true );
		main[i]->setupitem = groupItem;
		setup->AddItem( groupItem );
		
		// Now load in all the people in that group (backwards, to maintain order)
		plist = &(main[i]->names);
		for( unsigned int j = (*plist).Count(); j > 0; --j ) {
			personItem = new PeopleSetupItem( (*plist)[j-1]->name, false );
			(*plist)[j-1]->setupitem = personItem;
			setup->AddUnder( personItem, groupItem );
		}
	}

	lock->Unlock();
}

//-------------------------------------------------------------------------

PersonInfo* PeopleList::AddPerson( BString user, BString group, bool commitNow, bool check )
{
	lock->Lock();
	GroupInfo* theGroup;
	PersonInfo* pi;
	
	#ifdef PLIST_DEBUG
	printf( "attempting to add person %s in group %s (proto %u), check: %s\n", person, group, proto, check ? "yes" : "no" );
	#endif
	
	// if we are supposed to check if the person is already there, do it
	if( check ) {
		if( FindPerson( user ) ) {
			lock->Unlock();
			return NULL;		
		}		
	}
	
	// Whew. Done. Now find the group, and make it if it isn't there.
	theGroup = FindGroup( group, true );
	
	#ifdef PLIST_DEBUG
	printf( "adding person %s in group %s (proto %u)\n", person, group, proto );
	#endif
	
	// Set some information
	pi = new PersonInfo;
	pi->item = NULL;
	pi->warningLevel = 0;
	pi->setupitem = NULL;
	pi->parent = NULL;
	pi->status = BS_OFFLINE;
	pi->name = user;

	// Attach it to the list
	Attach( pi, theGroup );
	
	// add this person to the commit list
	AddCommitListItem( user, true );
	
	// if we're supposed to commit right away, do that
	if( commitNow )
		Commit();
	
	lock->Unlock();
	return pi;
}

//-------------------------------------------------------------------------

void PeopleList::RemovePerson( BString user, bool commitNow ) {

	lock->Lock();
	PersonInfo* pi;
	unsigned int groupIndex = 0;
	
	// Find the person in the list, and if successful, remove it
	pi = FindPerson( user, &groupIndex );
	if( pi )
		RemovePerson( pi, groupIndex, commitNow );
	
	lock->Unlock();
}

//-------------------------------------------------------------------------

void PeopleList::RemovePerson( PersonInfo* person, unsigned int groupIndex, bool commitNow ) {

	lock->Lock();
	
	#ifdef PLIST_DEBUG
	printf( "removing person %s...\n", person->name );
	#endif
	
	// detach the person from the list
	Detach( person, groupIndex );
	
	// Add the person to the commit list
	AddCommitListItem( person->name, false );
	
	// if we're supposed to commit right away, do that
	if( commitNow )
		Commit();	
	
	// delete the sucker
	delete person;
	lock->Unlock();
}

//-------------------------------------------------------------------------

void PeopleList::SetPersonStatus( BString user, unsigned int status, int32 wLevel, unsigned int changeTo )
{
	lock->Lock();
	PersonInfo* setPerson;

	// Try and find that person and leave if it fails
	setPerson = FindPerson( user );
	if( !setPerson ) {
		#ifdef PLIST_DEBUG
		printf( "person %u:%s not found! leaving.\n", proto, person );
		#endif
		lock->Unlock();
		return;
	}

	// call the function that does the real work
	SetPersonStatus( setPerson, status, wLevel, changeTo );	
	lock->Unlock();
}

//-------------------------------------------------------------------------

void PeopleList::SetPersonStatus( PersonInfo* person, unsigned int status, int32 wLevel, unsigned int changeTo )
{
	lock->Lock();
	
	bool doRefresh = false;

	#ifdef PLIST_DEBUG
	printf( "changing status of %s\n", person->name );
	#endif

	// Put them in the "online" list, if needed	
	if( !isOnline(person->status) && isOnline(status) && FixChangesForStatus(person,status) ) {
		SetOnline( person, status );
		changeTo = status;
		sounds->PlaySound( WS_ENTER );
	}
	
	// Take them OFF the "online" list, if needed
	else if( isOnline(person->status) && !isOnline(status) && FixChangesForStatus(person,status) ) {
		windows->GetBuddyList()->Lock();
		person->item->SetStatus( BS_LEAVING );
		windows->GetBuddyList()->Unlock();
		doRefresh = true;
		changeTo = BS_BYEBYE;
		sounds->PlaySound( WS_EXIT );
	}
	
	// they ready to finally leave now?
	else if( status == BS_BYEBYE ) {
		SetOffline( person );
		status = person->status;		// don't actually change the status
	}
	
	// update stuff
	else if( display && person->item ) {
		if( FixChangesForStatus( person, status ) ) {
			windows->GetBuddyList()->Lock();
			person->item->SetStatus( status );
			windows->GetBuddyList()->Unlock();
			doRefresh = true;		
		} else
			changeTo = 0;
	}
	
	// set the warning level if needed
	if( wLevel != -1 && display && person->item ) {
		if( wLevel != (int32)person->warningLevel ) {
			person->warningLevel = wLevel;
			windows->GetBuddyList()->Lock();
			person->item->SetWarningLevel( (unsigned int)wLevel );
			windows->GetBuddyList()->Unlock();
			doRefresh = true;
		}
	}
	
	// need to refresh the display?
	if( doRefresh ) {
		windows->GetBuddyList()->Lock();
		display->InvalidateItem( display->IndexOf(person->item) );
		windows->GetBuddyList()->Unlock();
	}
	
	// If this is going to require a change, set that up
	if( changeTo )
		SetupChange( person, changeTo );
		
	// now actually set their status
	person->status = status;		
	
	lock->Unlock();
}

//-------------------------------------------------------------------------

bool PeopleList::FixChangesForStatus( PersonInfo* person, unsigned int status ) {
	
	lock->Lock();
	
	// Go through and make sure that there isn't already a change record set up
	for( unsigned int i = 0; i < changeList.Count(); ++i ) {

		// found one... make sure it works with the existing change
		if( changeList[i].person == person && changeList[i].Counter != 0 ) {

			printf( "About to test... (%d)  (%d)\n", status, changeList[i].ChangeTo );
			
			// Is this person newly active, and we're just receiving a status update?
			if( status == BS_ACTIVE && changeList[i].ChangeTo == BS_ACTIVE ) {
				printf( "Test Completed! (1)...\n" );
				lock->Unlock();
				return false;
			}
			
			// Did this person just enter, and went offline right away?
			if( status == BS_OFFLINE && changeList[i].ChangeTo == BS_ACTIVE ) {
				printf( "Test Completed! (2)...\n" );
				changeList.Delete(i);
				lock->Unlock();
				return true;
			}
			
			// Did this person just go offline, and came right back in?
			if( (status == BS_ACTIVE || status == BS_IDLE) && changeList[i].ChangeTo == BS_BYEBYE ) {
				printf( "Test Completed! (3)...\n" );
				changeList.Delete(i);
				lock->Unlock();
				return true;
			}
		}
	}	
	
	lock->Unlock();
	return true;
}

//-------------------------------------------------------------------------

void PeopleList::ChangePerson( BString oldUser, BString newUser, BString newgroup )
{
	lock->Lock();
	PersonInfo* oldobject;
	unsigned int groupIndex;
	
	// Find the old person's object
	oldobject = FindPerson( oldUser, &groupIndex );
	if( !oldobject ) {
		lock->Unlock();
		return;
	}
	
	// Call the function that does the real work
	ChangePerson( oldobject, groupIndex, newUser, newgroup );
	
	lock->Unlock();		
}

//-------------------------------------------------------------------------

void PeopleList::ChangePerson( PersonInfo* oldperson, unsigned int groupIndex, BString newperson, BString newgroup )
{
	#ifdef PLIST_DEBUG
	printf( "changing %s (%s) to %s (%s)\n", oldperson->name.String(), oldperson->parent->group.String(), newperson.String(), newgroup.String() );
	#endif

	lock->Lock();
	bool onlyGroupChange = false;
	PeopleSetupItem* removeSetupItem;
	PeopleItem* removeItem;

	// Get the display items	
	removeItem = oldperson->item;
	removeSetupItem = oldperson->setupitem;
	
	// Check to see what, if anything, has changed
	//if( strcasecmp(oldperson->name, newperson) == 0 && oldperson->proto == newproto ) {
	if( CompareScreenNames(const_cast<char*>(oldperson->name.String()), const_cast<char*>(newperson.String())) ) {
	
		// if nothing has changed, we can return
		if( oldperson->parent->group == newgroup ) {
			lock->Unlock();
			return;
		}		
		// ONLY the group has changed... we can skip a bunch of steps
		else onlyGroupChange = true;
	}
	
	// change the old name in the list
	if( !onlyGroupChange ) {
	
		// check to see if the person is already there
		if( FindPerson( newperson ) ) {
			#ifdef PLIST_DEBUG
			printf( "That person is already there!\n" );
			#endif
			
			Say( "That person is already there!" );
			lock->Unlock();
			return;
		}
		
		// Remove the old person, and add the new one
		RemovePerson( oldperson, groupIndex );
		AddPerson( newperson, newgroup );
	}
	
	// Only the group was changed; everything else stays the same
	else {
	
		// Get the group item
		GroupInfo* groupItem = FindGroup( newgroup );
		if( !groupItem ) {
			lock->Unlock();
			return;
		}

		// Move the sucker
		Detach( oldperson, groupIndex );
		Attach( oldperson, groupItem );

		// correct the group numbers and display stuff
		if( setup && removeSetupItem && groupItem->setupitem ) {
			windows->GetBuddyList()->Lock();
			setup->MoveItem( setup->IndexOf(removeSetupItem), setup->IndexOf(groupItem->setupitem) + 1 );
			windows->GetBuddyList()->Unlock();
		}
	}
	
	lock->Unlock();		
}


//-------------------------------------------------------------------------

void PeopleList::CorrectGroupNumbers( GroupInfo* group, unsigned int status, bool add ) {

	// adjust the group's numbers
	if( add ) {
		group->total++;
		if( isOnline(status) )
			group->on++;
	} else {
		group->total--;
		if( isOnline(status) )
			group->on--;
	}
		
	// adjust the display, if needed
	if( display && group->item ) {
		windows->GetBuddyList()->Lock();
		group->item->SetNumbers( group->on, group->total );
		display->InvalidateItem( display->IndexOf(group->item) );
		windows->GetBuddyList()->Unlock();
	}
}

//-------------------------------------------------------------------------

// groupindex, if not null, returns the index of the person within the group
PersonInfo* PeopleList::FindPerson( BString userName, unsigned int* groupIndex )
{	
	lock->Lock();
	PersonInfo* pi = NULL;
	GenList<PersonInfo*>* plist;
	if( groupIndex )
		*groupIndex = 0;
	
	// Find the person... could be in any group... sigh
	for( unsigned int i = 0; i < main.Count(); ++i ) {
		plist = &(main[i]->names);
		for( unsigned int j = 0; j < plist->Count(); ++j ) {
			//if( strcasecmp( (*plist)[j]->name, person ) == 0 && 
			if( CompareScreenNames( (char*)(*plist)[j]->name.String(), (char*)userName.String() ) )  {
				pi = (*plist)[j];
				if( groupIndex )
					*groupIndex = j;
				break;
			}			
		}
	}
	
	#ifdef PLIST_DEBUG
	printf( "FindPerson %s: %s\n", person, pi == NULL ? "no" : "yes" );
	#endif
	
	lock->Unlock();
	return pi;
}

//-------------------------------------------------------------------------

GroupInfo* PeopleList::FindGroup( BString group, bool make, unsigned int* index )
{	
	GroupInfo* target = NULL;
	lock->Lock();
	
	// Look through all the groups
	for( unsigned int i = 0; i < main.Count(); ++i )
		if( main[i]->group == group ) {
			target = main[i];
			if( index )
				*index = i;
			break;
		}
			
	// If it's not there, make it (if we're supposed to)
	if( !target && make )
		target = AddGroup( group );

	lock->Unlock();
	return target;	
}

//-------------------------------------------------------------------------

void PeopleList::AddCommitListItem( BString userName, bool add ) {	

	lock->Lock();
	
	// make a person capsule
	PersonCapsule added;
	added.name = userName;
	added.add = add;

	#ifdef PLIST_DEBUG	
	printf( "attempting to %s person %s:%u %s commit list\n", add?"add":"remove", name, proto, add?"to":"from" );
	#endif
	
	// Go through and look for a duplicate
	for( unsigned int i = 0; i < commitList.Count(); ++i ) {
	
		// whoa... found one!
		//if( strcasecmp(name, commitList[i].name) == 0 && commitList[i].proto == proto ) {
		if( CompareScreenNames( (char*)userName.String(), (char*)commitList[i].name.String()) ) {

			// if it's not the same op, delete it... add/remove cancel each other out
			if( commitList[i].add != add ) {
				if( commitList[i].add )
					--commitAdds;
				commitList.Delete(i);				
			}
			
			// Otherwise, ignore it. No real point in adding/removing a name twice.
			lock->Unlock();
			return;
		}				
	}

	// should be ok to add it at this point... do so
	// adds go at the front of the list... removes go at the back
	if( add ) {
		commitList.Insert( added, 0 );
		++commitAdds;
	} else
		commitList.Push( added );

	lock->Unlock();
}

//-------------------------------------------------------------------------

void PeopleList::DispatchCommitList( GenList<PersonCapsule>& sendList, bool add ) {

	lock->Lock();
	BString sendString;
	BMessage* commitMsg;
	
	// add everyone in the list onto the send string
	for( unsigned int i = 0; i < sendList.Count(); ++i ) {
		if( i != 0 )
			sendString.Append( char(9), 1 );
		sendString.Append( sendList[i].name );
	}
	
	// make and send a BMessage for the commit stuff
	commitMsg = new BMessage(BEAIM_PEOPLELIST_COMMIT);
	commitMsg->AddBool( "add", add );
	commitMsg->AddString( "list", sendString.String() );
	PostAppMessage( commitMsg );
	
	printf( "  dispatching %s list\n", add?"add":"remove" );
	printf( "---->|%s|\n\n", sendString.String() );

	lock->Unlock();
}

//-------------------------------------------------------------------------

void PeopleList::ClearCommitList() {
	commitList.Clear();
	commitAdds = 0;
}

//-------------------------------------------------------------------------

void PeopleList::Commit()
{
	lock->Lock();
	GenList<PersonCapsule> committed;

	#ifdef PLIST_DEBUG
	printf( "total commit list size: %u\n", commitList.Count() );
	printf( "commit list adds: %u\n", commitAdds );
	printf( "commit list removes: %u\n", commitList.Count() - commitAdds );
	#endif

	// Make an add list for that protocol
	committed.Clear();
	for( unsigned int j = 0; j < commitAdds; ++j ) {
		#ifdef PLIST_DEBUG
		printf( "  add: %s\n", commitList[j].name );
		#endif
		committed.Add( commitList[j] );
	}
	if( committed.Count() )
		DispatchCommitList( committed, true );

	// Make a remove list for that protocol
	committed.Clear();
	for( unsigned int j = commitAdds; j < commitList.Count(); ++j ) {
		#ifdef PLIST_DEBUG
		printf( "  remove: %s\n", commitList[j].name );
		#endif
		committed.Add( commitList[j] );
	}
	if( committed.Count() )
		DispatchCommitList( committed, false );	

	// reset the commit list variables
	commitAdds = 0;
	commitList.Clear();
	lock->Unlock();	
}

//-------------------------------------------------------------------------

void PeopleList::SetOnline( BString userName, unsigned int status ) {

	lock->Lock();
	PersonInfo* thePerson = FindPerson( userName );
	if( thePerson )
		SetOnline( thePerson, status );
	lock->Unlock();
}

//-------------------------------------------------------------------------

void PeopleList::SetOnline( PersonInfo* person, unsigned int status ) {

	lock->Lock();
	GroupInfo* group = person->parent;
	PeopleItem* newPerson;
	
	#ifdef PLIST_DEBUG
	printf( "setting online: %s\n", person->name );
	#endif
	
	// update the display
	if( display ) {
		windows->GetBuddyList()->Lock();
		if( person->item )
			person->item->SetStatus( status | BS_ENTERING );	
		else {
			group->on++;
			newPerson = new PeopleItem( person->name );
			newPerson->SetStatus( status | BS_ENTERING );
			display->AddUnder( newPerson, group->item );
			person->item = newPerson;
		}

		group->item->SetNumbers( group->on, group->total );
		display->InvalidateItem( display->IndexOf(group->item) );
		display->InvalidateItem( display->IndexOf(person->item) );
		windows->GetBuddyList()->Unlock();
	}
	
	lock->Unlock();
}

//-------------------------------------------------------------------------

void PeopleList::SetOffline( BString userName ) {

	lock->Lock();
	PersonInfo* thePerson = FindPerson( userName );
	if( thePerson )
		SetOffline( thePerson );
	lock->Unlock();
}

//-------------------------------------------------------------------------

void PeopleList::SetOffline( PersonInfo* person ) {

	lock->Lock();
	GroupInfo* group = person->parent;
	
	// Increment the online group count
	group->on--;
		
	// update the display
	if( display ) {
		windows->GetBuddyList()->Lock();
		group->item->SetNumbers( group->on, group->total );
		display->InvalidateItem( display->IndexOf(group->item) );
		display->RemoveItem( person->item );
		person->item = NULL;
		windows->GetBuddyList()->Unlock();
	}
	
	lock->Unlock();
}

//-------------------------------------------------------------------------

void PeopleList::GetGroupInfo( BString group, unsigned int& online, unsigned int& total ) {

	lock->Lock();
	total = 0;
	online = 0;
	
	// Get the group and return if not there
	GroupInfo* gp = FindGroup( group );
	if( !gp ) {
		lock->Unlock();
		return;
	}
	
	total = gp->total;
	online = gp->on;
	
	#ifdef PLIST_DEBUG
	printf( "%s: (%u/%u)\n", group, online, total );
	#endif
	
	lock->Unlock();	
}

//-------------------------------------------------------------------------

void PeopleList::ChangeGroup( BString oldname, BString newname ) {
			
	lock->Lock();
	unsigned int groupPos;
	GroupInfo* group;
	PeopleSetupItem* psi;
	GroupItem* gi;
	
	// Find the group
	group = FindGroup( oldname, &groupPos );
	if( !group ) {
		lock->Unlock();
		return;
	}
	
	// Check to see if the new group already exists
	if( FindGroup( newname ) ) {
		#ifdef PLIST_DEBUG
		printf( "group already exists!\n" );
		#endif
		
		Say( "That group already exists!" );
		lock->Unlock();
		return;
	}
	
	// change the name
	group->group = newname;
	psi = group->setupitem;
	gi = group->item;
		
	// update the displays
	if( setup && psi ) {
		psi->SetUserName( newname );
		setup->InvalidateItem( setup->IndexOf( psi ) );
	}
	if( display && gi ) {
		gi->SetGroupName( newname );
		display->InvalidateItem( display->IndexOf( gi ) );
	}
	
	lock->Unlock();			
}

//-------------------------------------------------------------------------

unsigned int PeopleList::CountPeople() {
//	return main.Count();

	lock->Lock();
	unsigned int count = 0;
	
	for( unsigned int i = 0; i < main.Count(); ++i )
		count += main[i]->names.Count();
		
	lock->Unlock();
	return count;
}

//-------------------------------------------------------------------------		
		
void PeopleList::GetPersonInfo( BString userName, PersonInfo*& ret ) {

	lock->Lock();
	ret = FindPerson( userName );
	lock->Unlock();	
}
		
//-------------------------------------------------------------------------

void PeopleList::GetUserList( BString& list ) {

	lock->Lock();
	unsigned int count = 0;
	GenList<PersonInfo*>* plist;
	
	// Go through and count the number of people for this protocol
	for( unsigned int i = 0; i < main.Count(); ++i ) {
		plist = &(main[i]->names);
		count += (*plist).Count();
	}
	
	#ifdef PLIST_DEBUG
	printf( "list proto count: %u\n", count );
	#endif
	
	if( !count ) {
		list = NULL;
		lock->Unlock();
		return;
	}

	// Go through and get the names for this protocol
	count = 0;
	for( unsigned int i = 0; i < main.Count(); ++i ) {
		plist = &(main[i]->names);
		for( unsigned int j = 0; j < (*plist).Count(); ++j ) {
			if( count++ )
				list.Append( char(9), 1 );
			list.Append( (*plist)[j]->name );
		}
	}	
	
	lock->Unlock();
}

//-------------------------------------------------------------------------

GroupInfo* PeopleList::AddGroup( BString group, bool check ) {

	lock->Lock();
	GroupItem* groupItem;
	GroupInfo* newGroup;
	PeopleSetupItem* groupSetupItem;
	
	// if we are supposed to, check and see if this group already exists
	if( check && FindGroup( group, false ) ) {
		lock->Unlock();
		return NULL;
	}
		
	#ifdef PLIST_DEBUG
	printf( "---> adding group %s\n", group );
	#endif
		
	// OK, it doesn't exist. Add it.
	newGroup = new GroupInfo;
	newGroup->on = newGroup->total = 0;
	newGroup->item = NULL;
	newGroup->setupitem = NULL;
	newGroup->group = group;
	
	// Add it to the displays
	if( setup && display ) {
		BWindow* bList = windows->GetBuddyList();
		if( bList )
			bList->Lock();
		groupItem = new GroupItem( group, 0, 0 );
		groupSetupItem = new PeopleSetupItem( group, true );
		newGroup->item = groupItem;
		newGroup->setupitem = groupSetupItem;
		display->AddItem( groupItem );
		setup->AddItem( groupSetupItem );
		if( bList )
			bList->Unlock();
	}

	// add it
	main.Add( newGroup );
	
	lock->Unlock();
	return newGroup;
}

//-------------------------------------------------------------------------

void PeopleList::RemoveGroup( BString group ) {
		
	lock->Lock();
	unsigned int index = 0;
	GroupInfo* grp;
	
	// Find the group object and its index
	grp = FindGroup( group, false, &index );

	#ifdef PLIST_DEBUG
	printf( "RemoveGroup: %s index: %u\n", group, index );
	#endif
	
	// Remove all the entries in the group
	for( unsigned int i = grp->names.Count(); i > 0; --i )
		RemovePerson( grp->names[i-1], i-1 );

	// Remove the group items from the displays
	if( display && grp->item )
		display->RemoveItem( grp->item );
	if( setup && grp->setupitem )
		setup->RemoveItem( grp->setupitem );		

	// Remove the group itself
	main.Delete( index );
	
	lock->Unlock();			
}

//-------------------------------------------------------------------------

unsigned int PeopleList::CountGroups() {
	
	lock->Lock();
	unsigned int count = main.Count();
	lock->Unlock();
	return count;
}
		
//-------------------------------------------------------------------------

void PeopleList::GetGroup( unsigned int i, BString& group ) {

	lock->Lock();
	group = main[i]->group;
	lock->Unlock();	
}

//-------------------------------------------------------------------------

void PeopleList::Debug() {
	
/*	lock->Lock();
	#ifdef PLIST_DEBUG
	printf( "\n\n" );
	#endif
	GenList<PersonInfo*>* plist;
	char debug[512];
	
	for( unsigned int i = 0; i < main.Count(); ++i ) {
		
		plist = &(main[i]->names);
		for( unsigned int j = 0; j < (*plist).Count(); ++j ) {
			sprintf( debug, "%s: %s (%u)", main[i]->group, (*plist)[j]->name, (*plist)[j]->proto );
			#ifdef PLIST_DEBUG
			printf( "%s\n", debug );
			#endif
		}
	}
	
	lock->Unlock();	*/
}

//-------------------------------------------------------------------------

void PeopleList::UpdateChangeList() {

	lock->Lock();

	// go through backwards to avoid corrupting array indices
	for( int32 i = changeList.Count() - 1; i >= 0; i-- ) {
	
		// decrement the time counter
		changeList[i].Counter--;

		if( changeList[i].Counter == 0 ) {
				
			// If the new status is BS_BYEBYE, it means to remove the buddy
			if( changeList[i].ChangeTo == BS_BYEBYE )
				display->RemoveItem( changeList[i].person->item );
			
			// Set the person's status
			SetPersonStatus( changeList[i].person, changeList[i].ChangeTo );

			// Delete the change from the changelist
			changeList.Delete(i);			
		}
	}
		
	lock->Unlock();	
}

//-------------------------------------------------------------------------

void PeopleList::SetupChange( BString userName, unsigned int changeTo ) {

	lock->Lock();
	PersonInfo* thePerson = FindPerson( userName );
	if( thePerson )
		SetupChange( thePerson, changeTo );
	lock->Unlock();
}

//-------------------------------------------------------------------------

void PeopleList::SetupChange( PersonInfo* person, unsigned int changeTo ) {

	// Set up a new changeList entry
	lock->Lock();
	PersonChange change;
	change.ChangeTo = changeTo;
	change.Counter = BUDDY_CHANGE_TIME;
	change.person = person;
 
	// Go through and make sure that there isn't already a change record set up
	for( unsigned int i = 0; i < changeList.Count(); ++i ) {

		// if there is, nuke it
		if( changeList[i].person == person ) {
			//changeList[i] = change;
			changeList.Delete(i);
			//lock->Unlock();
			return;
		}
	}

	// OK, no existing change record... so add a new one
	changeList.Add( change );	
	lock->Unlock();	
}

//-------------------------------------------------------------------------

bool PeopleList::isOnline( const uint32 status ) {

	lock->Lock();
	bool ret;
	
	switch( status ) {
	
		case BS_OFFLINE:
		case BS_LEAVING:
		case BS_BYEBYE:
			ret = false;
			break;
			
		default:
			ret = true;
			break;	
	}
	
	lock->Unlock();
	return ret;
}

//-------------------------------------------------------------------------

void PeopleList::Attach( PersonInfo* newPerson, BString group, bool make ) {

	lock->Lock();
	GroupInfo* theGroup = FindGroup( group, make );
	Attach( newPerson, theGroup );
	lock->Unlock();
}

//-------------------------------------------------------------------------

void PeopleList::Attach( PersonInfo* person, GroupInfo* group ) {

	lock->Lock();
	PeopleSetupItem* psi;
	PeopleItem* pi;
	
	#ifdef PLIST_DEBUG
	printf( "attaching %s to group %s\n", person->name.String(), group->group.String() );
	#endif

	// Add it to the list
	group->names.Add( person );
	person->parent = group;
	
	BWindow* bList = windows->GetBuddyList();
	
	// Add it to the setup display
	if( setup && group->setupitem && bList ) {
		bList->Lock();
		psi = new PeopleSetupItem( person->name, false );
		setup->AddUnder( psi, group->setupitem );
		person->setupitem = psi;
		bList->Unlock();
	}	

	// Add it to the regular display
	if( display && group->item && bList && isOnline(person->status) ) {
		bList->Lock();
		pi = new PeopleItem( person->name, person->status );
		display->AddUnder( pi, group->item );
		person->item = pi;
		bList->Unlock();
	}	

	// set the new group stuff
	CorrectGroupNumbers( group, person->status, true );
	
	lock->Unlock();
}
		
//-------------------------------------------------------------------------
		
void PeopleList::Detach( PersonInfo* person, unsigned groupIndex ) {

	lock->Lock();
	GroupInfo* group = person->parent;
	
	#ifdef PLIST_DEBUG
	printf( "detaching %s from group %s\n", person->name, person->parent->group );
	#endif

	// remove the person from the list
	group->names.Delete( groupIndex );
	person->parent = NULL;
	CorrectGroupNumbers( group, person->status, false );
	BWindow* bList = windows->GetBuddyList();
	
	// Remove it from the setup display
	if( setup && person->setupitem && bList ) {
		bList->Lock();
		setup->RemoveItem( person->setupitem );
		bList->Unlock();
	}
		
	// Remove it from the online display
	if( display && person->item ) {
		bList->Lock();
		display->RemoveItem( person->item );
		bList->Unlock();
	}

	lock->Unlock();
}

//-------------------------------------------------------------------------

bool PeopleList::GetFirstGroup( BString& name ) {

	lock->Lock();
	GroupInfo* gp;
	bool ret = main.First(gp);
	if( !ret ) {
		lock->Unlock();
		return false;
	}
	name = BString(gp->group);
	lock->Unlock();
	return true;
}

//-------------------------------------------------------------------------

bool PeopleList::GetNextGroup( BString& name ) {

	lock->Lock();
	GroupInfo* gp;
	bool ret = main.Next(gp);
	if( !ret ) {
		lock->Unlock();
		return false;
	}
	name = BString(gp->group);
	lock->Unlock();
	return true;
}

//-------------------------------------------------------------------------

bool PeopleList::GetFirstBuddyForCurrentGroup( BString& name ) {

	lock->Lock();
	GroupInfo* gp;
	PersonInfo* pi;
	bool ret = main.Current(gp);
	if( !ret ) {
		lock->Unlock();
		return false;
	}
	ret = gp->names.First(pi);
	if( !ret ) {
		lock->Unlock();
		return false;
	}
	name = BString(pi->name);
	lock->Unlock();
	return true;
}

//-------------------------------------------------------------------------

bool PeopleList::GetNextBuddyForCurrentGroup( BString& name ) {

	lock->Lock();
	GroupInfo* gp;
	PersonInfo* pi;
	bool ret = main.Current(gp);
	if( !ret ) {
		lock->Unlock();
		return false;
	}
	ret = gp->names.Next(pi);
	if( !ret ) {
		lock->Unlock();
		return false;
	}
	name = BString(pi->name);
	lock->Unlock();
	return true;
}

//=========================================================================

PeopleSetupItem::PeopleSetupItem( BString s, bool group ) {

	userName = s;
	isgroup = group;
}

//-------------------------------------------------------------------------

PeopleSetupItem::~PeopleSetupItem() {
}

//-------------------------------------------------------------------------

BString PeopleSetupItem::UserName() {
	return userName;
}

//-------------------------------------------------------------------------

void PeopleSetupItem::DrawItem( BView *owner, BRect frame, bool complete ) {

	rgb_color color;

	// Make the selection color (shouldn't be hardcoded!)
	rgb_color kHighlight; 
	kHighlight.red = kHighlight.blue = 222;
	kHighlight.green = 219;
	
	// Grab the owner's font, put it into boldface if it's expanded
	BFont baseFont;
	owner->GetFont( &baseFont );
	BFont ownerFont( baseFont );
	
	// does the background need to be redrawn?
	if( IsSelected() || complete ) {
		
		// pick the appropriate background color
        if( IsSelected() ) {
        	color = kHighlight;
        	owner->SetLowColor( 222, 219, 222 );
        } else {
			color = owner->ViewColor();
			owner->SetLowColor( 255, 255, 255 );
		}
	
		// draw the background		
		owner->SetHighColor(color);
		owner->FillRect(frame);
	}

	// set the font and draw the string
	owner->SetHighColor(0,0,0);
	owner->SetFont( &ownerFont );
	owner->MovePenTo(frame.left + 4, frame.bottom - 2);
	owner->DrawString( userName.String() ); 
	
	// reset the font and colors
	owner->SetFont( &baseFont );
	owner->SetLowColor( 255, 255, 255 );
}

//-------------------------------------------------------------------------

void PeopleSetupItem::SetUserName( BString u ) {
	userName = u;
}

//=========================================================================

GroupItem::GroupItem( BString n, unsigned int on, unsigned int tot ) {
	
	name = n;
	online = on;
	total = tot;
}

//-------------------------------------------------------------------------

GroupItem::~GroupItem() {
}

//-------------------------------------------------------------------------

BString GroupItem::GroupName() {
	return name;
}

//-------------------------------------------------------------------------

void GroupItem::DrawItem( BView *owner, BRect frame, bool complete ) {

	rgb_color color;
	
	// Make a string with the right stuff in it
	char* GroupTitle = (char*)malloc( name.Length() + 25 );
	sprintf( GroupTitle, "%s (%u/%u)", name.String(), online, total );

	// Make the selection color (shouldn't be hardcoded!)
	rgb_color kHighlight; 
	kHighlight.red = kHighlight.blue = 222;
	kHighlight.green = 219;
	
	// Grab the owner's font, put it into boldface if it's expanded
	BFont baseFont;
	owner->GetFont( &baseFont );
	BFont ownerFont( baseFont );
	if( IsExpanded() )
		ownerFont.SetFace( B_BOLD_FACE );
	
	// does the background need to be redrawn?
	if( IsSelected() || complete ) {
		
		// pick the appropriate background color
        if( IsSelected() ) {
        	color = kHighlight;
        	owner->SetLowColor( 222, 219, 222 );
        } else {
			color = owner->ViewColor();
			owner->SetLowColor( 255, 255, 255 );
		}
	
		// draw the background		
		owner->SetHighColor(color);
		owner->FillRect(frame);
	}

	// set the font and draw the string
	owner->SetHighColor(0,77,74);
	owner->SetFont( &ownerFont );
	owner->MovePenTo(frame.left + 4, frame.bottom - 2);
	owner->DrawString( GroupTitle ); 
	
	// reset the font and colors
	owner->SetFont( &baseFont );
	owner->SetLowColor( 255, 255, 255 );
}

//-------------------------------------------------------------------------

void GroupItem::SetGroupName( BString s ) {
	name = s;
}

//-------------------------------------------------------------------------

void GroupItem::SetNumbers( unsigned int on, unsigned int tot ) {

	online = on;
	total = tot;
}

//=========================================================================

PeopleItem::PeopleItem( BString u, int32 stat ) {

	user = u;
	status = stat;
	warningLevel = 0;
	if( awayIcon == NULL )
		GetBitmapFromResources( awayIcon, 3827 );
		
	MakeDisplayName();
}

//-------------------------------------------------------------------------

void PeopleItem::SetUserName( BString u ) {
	user = u;
	MakeDisplayName();
}

//-------------------------------------------------------------------------

void PeopleItem::MakeDisplayName() {

	displayname = user;
	if( warningLevel ) {
		char levelString[10];
		sprintf( levelString, "%u", warningLevel );
		displayname.Append( "  (" );
		displayname.Append( levelString );
		displayname.Append( "%)" );
	}
}

//-------------------------------------------------------------------------

void PeopleItem::SetWarningLevel( unsigned int lvl ) {
	warningLevel = lvl;
	MakeDisplayName();
}

//-------------------------------------------------------------------------

PeopleItem::~PeopleItem() {
}

//-------------------------------------------------------------------------

BString PeopleItem::UserName() {
	return user;
}

//-------------------------------------------------------------------------

void PeopleItem::SetStatus( unsigned int stat ) {
	status = stat;
}

//-------------------------------------------------------------------------

void PeopleItem::DrawItem( BView *owner, BRect frame, bool complete ) {

	rgb_color color;
	BRect awayRect = BRect(0,0,11,11);

	// Make the selection color (shouldn't be hardcoded!)
	rgb_color kHighlight; 
	kHighlight.red = kHighlight.blue = 222;
	kHighlight.green = 219;
	
	// Grab the owner's font, to be fiddled with if needed
	BFont baseFont;
	owner->GetFont( &baseFont );
	BFont ownerFont( baseFont );
	
	// does the background need to be redrawn?
	frame.left++;
	if( IsSelected() || complete ) {
		
		// pick the appropriate background color
        if( IsSelected() ) {
        	color = kHighlight;
        	owner->SetLowColor( 222, 219, 222 );
        } else {
			color = owner->ViewColor();
			owner->SetLowColor( 255, 255, 255 );
		}
	
		// draw the background		
		owner->SetHighColor(color);
		owner->FillRect(frame);
	}
	
	// Set colors and font attributes based on buddy status
	if( status & BS_ENTERING )
		ownerFont.SetFace( B_BOLD_FACE );
	if( status & BS_LEAVING )
		ownerFont.SetFace( B_ITALIC_FACE );
	if( status & BS_IDLE )
		owner->SetHighColor(105,106,105);
	else
		owner->SetHighColor(0,0,0);


	// set the font and draw the string
	owner->SetFont( &ownerFont );
	owner->MovePenTo(frame.left + 5, frame.bottom - 2);
	owner->DrawString( displayname.String() );
	
	// draw the bitmap, if away
	if( status & BS_AWAY ) {
		awayRect.OffsetTo( frame.left - 13, frame.top + 3 );
		owner->DrawBitmap( awayIcon, awayRect );
	}

	// reset the font and colors
	owner->SetFont( &baseFont );
	owner->SetLowColor( 255, 255, 255 );
	owner->SetHighColor( 0, 0, 0 );
}

//-------------------------------------------------------------------------